// This file is auto-generated. See "generate_kernels.sh"
#include "../kernel_backward.h"
INSTANTIATE_ATTENTION_KERNEL_BACKWARD_SM50(float, true);
INSTANTIATE_ATTENTION_KERNEL_BACKWARD_SM70(float, true);
INSTANTIATE_ATTENTION_KERNEL_BACKWARD_SM75(float, true);
INSTANTIATE_ATTENTION_KERNEL_BACKWARD_SM80(float, true);
